"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = exports.PAGE_UNLOADED_DURING_EXECUTION_ERROR_MESSAGE = void 0;

var _cssTree = _interopRequireDefault(require("css-tree"));

var _debug = _interopRequireDefault(require("debug"));

var _pruneNonCriticalSelectors = _interopRequireDefault(require("./browser-sandbox/pruneNonCriticalSelectors"));

var _replacePageCss = _interopRequireDefault(require("./browser-sandbox/replacePageCss"));

var _insertStyles = _interopRequireDefault(require("./browser-sandbox/insertStyles"));

var _criticalHtmlImages = _interopRequireDefault(require("./browser-sandbox/criticalHtmlImages"));

var _postformatting = _interopRequireDefault(require("./postformatting"));

var _criticalCssImages = _interopRequireDefault(require("./postformatting/criticalCssImages"));

var _selectorsProfile = _interopRequireDefault(require("./selectors-profile"));

var _nonMatchingMediaQueryRemover = _interopRequireDefault(require("./non-matching-media-query-remover"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance"); }

function _iterableToArray(iter) { if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i]; return arr2; } }

const debuglog = (0, _debug.default)('penthouse:core');
const PUPPETEER_PAGE_UNLOADED_DURING_EXECUTION_ERROR_REGEX = /(Cannot find context with specified id|Execution context was destroyed)/;
const PAGE_UNLOADED_DURING_EXECUTION_ERROR_MESSAGE = 'PAGE_UNLOADED_DURING_EXECUTION: Critical css generation script could not be executed.\n\nThis can happen if Penthouse was killed during execution, OR otherwise most commonly if the page navigates away after load, via setting window.location, meta tag refresh directive or similar. For the critical css generation to work the loaded page must stay: remove any redirects or move them to the server. You can also disable them on your end just for the critical css generation, for example via a query parameter.';
exports.PAGE_UNLOADED_DURING_EXECUTION_ERROR_MESSAGE = PAGE_UNLOADED_DURING_EXECUTION_ERROR_MESSAGE;

function loadPage({
  page,
  url,
  pageLoadSkipTimeout,
  allowedResponseCode,
  pageTargetInfoId
}) {
  debuglog(`page load START (${pageTargetInfoId}, ${url})`);
  let timeoutId;
  const loadPagePromise = page.goto(url).then(response => {
    clearTimeout(timeoutId);

    if (typeof allowedResponseCode !== 'undefined') {
      checkResponseStatus(allowedResponseCode, response, timeoutId);
    }

    return page.content().then(html => {
      if (!html) {
        throw new Error('Server did not return any html. This can happen due to encoding issues, for example with gzip');
      }

      debuglog(`page load DONE (${pageTargetInfoId}, ${url})`);
    });
  }).catch(err => {
    debuglog('caught page load error, clearing timeout', err);
    clearTimeout(timeoutId);
    throw err;
  });

  if (!pageLoadSkipTimeout) {
    return loadPagePromise;
  }

  return Promise.race([loadPagePromise, new Promise(resolve => {
    // _abort_ page load after X time,
    // in order to deal with spammy pages that keep sending non-critical requests
    // (tracking etc), which would otherwise never load.
    timeoutId = setTimeout(() => {
      debuglog(`page load waiting ABORTED after ${pageLoadSkipTimeout / 1000}s. (${pageTargetInfoId}, ${url}`);
      resolve();
    }, pageLoadSkipTimeout);
  })]);
}

function checkResponseStatus(allowedResponseCode, response, timeoutId) {
  var errorMessage;

  if (typeof allowedResponseCode === 'number' && response.status() !== allowedResponseCode) {
    errorMessage = `Server response status ${response.status()} isn't matching allowedResponseCode: ${allowedResponseCode}.`;
  } else if (typeof allowedResponseCode === 'object' && allowedResponseCode.constructor.name === 'RegExp' && !response.status().toString().match(allowedResponseCode)) {
    errorMessage = `Server response status ${response.status()} isn't matching allowedResponseCode: ${allowedResponseCode.toString()}.`;
  } else if (typeof allowedResponseCode === 'function' && !allowedResponseCode.call(this, response)) {
    errorMessage = `Server response status ${response.status()} isn't matching allowedResponseCode.`;
  }

  if (errorMessage) {
    clearTimeout(timeoutId);
    throw new Error(errorMessage);
  }
}

async function astFromCss({
  cssString,
  strict
}) {
  // breaks puppeteer
  const css = cssString.replace(/￿/g, '\f042');
  const parsingErrors = [];
  debuglog('parse ast START');
  const csstreeOptions = strict && {
    onParseError: error => parsingErrors.push(error.formattedMessage)
  };

  const ast = _cssTree.default.parse(css, csstreeOptions);

  debuglog(`parse ast DONE ${strict ? `in strict mode (with ${parsingErrors.length} errors)` : ''}`);

  if (parsingErrors.length && strict === true) {
    // NOTE: only informing about first error, even if there were more than one.
    const parsingErrorMessage = parsingErrors[0];
    throw new Error(`AST parser (css-tree) found ${parsingErrors.length} errors in CSS.
      Breaking because in strict mode.
      The first error was:
      ` + parsingErrorMessage);
  }

  return ast;
}

async function preparePage({
  page,
  pagePromise,
  url,
  width,
  height,
  cookies,
  userAgent,
  customPageHeaders,
  blockJSRequests,
  cleanupAndExit,
  getHasExited
}) {
  let reusedPage;
  let pageTargetInfoId;

  try {
    const pagePromiseResult = await pagePromise;
    page = pagePromiseResult.page;
    reusedPage = pagePromiseResult.reused; // only used for debugging

    pageTargetInfoId = page && page._target._targetInfo.targetId;
  } catch (e) {
    debuglog('unexpected: could not get an open browser page' + e);
    return;
  } // we already exited while page was opening, stop execution
  // (strict mode ast css parsing errors)


  if (getHasExited()) {
    return;
  }

  debuglog(`open page ready in browser (${pageTargetInfoId}, ${url})`); // We set the viewport size in the browser when it launches,
  // and then re-use it for each page (to avoid extra work).
  // Only if later pages use a different viewport size do we need to
  // update it here.

  let setViewportPromise = Promise.resolve();
  const currentViewport = page.viewport();

  if (currentViewport.width !== width || currentViewport.height !== height) {
    setViewportPromise = page.setViewport({
      width,
      height
    }).then(() => debuglog('viewport size updated'));
  }

  const setUserAgentPromise = page.setUserAgent(userAgent).then(() => debuglog('userAgent set'));
  let setCustomPageHeadersPromise = Promise.resolve();

  if (customPageHeaders && Object.keys(customPageHeaders).length) {
    try {
      setCustomPageHeadersPromise = page.setExtraHTTPHeaders(customPageHeaders).then(() => debuglog('customPageHeaders set:' + JSON.stringify(customPageHeaders)));
    } catch (e) {
      debuglog(`failed setting extra http headers: (${pageTargetInfoId}, ${url})` + e);
    }
  }

  let setCookiesPromise = Promise.resolve();

  if (cookies) {
    try {
      var _page;

      setCookiesPromise = (_page = page).setCookie.apply(_page, _toConsumableArray(cookies)).then(() => debuglog('cookie(s) set: ' + JSON.stringify(cookies)));
    } catch (e) {
      debuglog(`failed to set cookies: (${pageTargetInfoId}, ${url})` + e);
    }
  } // NOTE: with JS disabled we cannot use JS timers inside page.evaluate
  // (setTimeout, setInterval), however requestAnimationFrame works.
  // NOTE: we are not explicitly blocking JS request; disabling JS takes care of that for us.


  const blockJSRequestsPromise = page.setJavaScriptEnabled(!blockJSRequests).then(() => {
    debuglog(`setJavaScriptEnabled${blockJSRequests ? '=false' : ''} DONE`);
  }); // assumes the page was already configured from previous call!

  if (reusedPage) {
    return Promise.all([setViewportPromise, setUserAgentPromise, setCustomPageHeadersPromise, setCookiesPromise, blockJSRequestsPromise]).then(() => {
      debuglog(`preparePage DONE (${pageTargetInfoId}, ${url})`);
      return page;
    });
  } // disable Puppeteer navigation timeouts;
  // Penthouse tracks these internally instead.


  page.setDefaultNavigationTimeout(0);
  page.on('error', error => {
    debuglog('page error: ' + error);
    cleanupAndExit({
      error
    });
  });
  page.on('console', msg => {
    const text = msg.text ? typeof msg.text === 'function' ? msg.text() : msg.text : msg; // pass through log messages
    // - the ones sent by penthouse for debugging has 'debug: ' prefix.

    if (/^debug: /.test(text)) {
      debuglog(text.replace(/^debug: /, ''));
    }
  });
  debuglog('page event listeners set');
  return Promise.all([setViewportPromise, setUserAgentPromise, setCustomPageHeadersPromise, setCookiesPromise]).then(() => {
    debuglog('preparePage DONE');
    return page;
  });
}

async function grabPageScreenshot({
  type,
  page,
  screenshots,
  screenshotExtension,
  debuglog,
  pageTargetInfoId,
  url
}) {
  const path = screenshots.basePath + `-${type}` + screenshotExtension;
  debuglog(`take ${type} screenshot START (${pageTargetInfoId}, ${url})`);
  return page.screenshot({ ...screenshots,
    path
  }).then(() => debuglog(`take ${type} screenshot DONE (${pageTargetInfoId}, ${url}), path: ${path}`)).catch(err => {
    debuglog(`take ${type} screenshot ERROR (${pageTargetInfoId}, ${url}) ⚠️ - ${err}`);
    throw err;
  });
}

async function pruneNonCriticalCssLauncher({
  pagePromise,
  url,
  cssString,
  width,
  height,
  formatCriticalCss,
  forceInclude,
  forceExclude,
  strict,
  userAgent,
  renderWaitTime,
  timeout,
  pageLoadSkipTimeout,
  blockJSRequests,
  customPageHeaders,
  cookies,
  screenshots,
  propertiesToRemove,
  maxEmbeddedBase64Length,
  keepLargerMediaQueries,
  detectCriticalImages,
  maxElementsToCheckPerSelector,
  allowedResponseCode,
  prepareDomFn
}) {
  let _hasExited = false; // hacky to get around _hasExited only available in the scope of this function

  const getHasExited = () => _hasExited;

  const takeScreenshots = screenshots && screenshots.basePath;
  const screenshotExtension = takeScreenshots && screenshots.type === 'jpeg' ? '.jpg' : '.png'; // NOTE: would need a refactor to killTimeout logic to be able to remove promise here.

  /* eslint-disable no-async-promise-executor */

  return new Promise(async (resolve, reject) => {
    /* eslint-enable no-async-promise-executor */
    debuglog('Penthouse core start');
    let page = null;
    let pageTargetInfoId; // only used fore debugging

    let killTimeoutId = null;

    async function cleanupAndExit({
      error,
      returnValue
    }) {
      if (_hasExited) {
        return;
      }

      debuglog(`cleanupAndExit (${pageTargetInfoId}, ${url})${error ? ` ⚠️ - ${error}` : ''} `);
      _hasExited = true;
      clearTimeout(killTimeoutId);

      if (error) {
        return reject(error);
      }

      if (page) {
        const resetPromises = []; // reset page headers and cookies,
        // since we re-use the page

        if (customPageHeaders && Object.keys(customPageHeaders).length) {
          try {
            resetPromises.push(page.setExtraHTTPHeaders({}).then(() => debuglog('customPageHeaders reset')));
          } catch (e) {
            debuglog(`failed resetting extra http headers: (${pageTargetInfoId}, ${url})` + e);
          }
        } // reset cookies


        if (cookies && cookies.length) {
          try {
            var _page2;

            resetPromises.push((_page2 = page).deleteCookie.apply(_page2, _toConsumableArray(cookies)).then(() => debuglog('cookie(s) reset: ')));
          } catch (e) {
            debuglog(`failed to reset cookies: (${pageTargetInfoId}, ${url})` + e);
          }
        }

        await Promise.all(resetPromises);
      }

      return resolve(returnValue);
    }

    killTimeoutId = setTimeout(() => {
      cleanupAndExit({
        error: new Error(`Penthouse timed out after ${timeout / 1000}s. (${pageTargetInfoId}, ${url})`)
      });
    }, timeout); // 1. start preparing a browser page (tab) [NOT BLOCKING]

    const updatedPagePromise = preparePage({
      page,
      pagePromise,
      url,
      width,
      height,
      userAgent,
      cookies,
      customPageHeaders,
      blockJSRequests,
      cleanupAndExit,
      getHasExited
    }).then(page => {
      // only used for debugging
      pageTargetInfoId = page && page._target._targetInfo.targetId;
      return page;
    }).catch(err => {
      debuglog(`preparePage crashed ${err} (${pageTargetInfoId}, ${url})`);
      cleanupAndExit({
        error: 'Could not open page in browser'
      });
    }); // 2. parse ast

    let ast = null;
    const astPromise = astFromCss({
      cssString,
      strict
    }).then(ast => {
      // 3. Further process the ast
      // Strip out non matching media queries.
      // Need to be done before buildSelectorProfile;
      // (very fast but could be done together/in parallel in future)
      (0, _nonMatchingMediaQueryRemover.default)(ast, width, height, keepLargerMediaQueries);
      debuglog(`stripped out non matching media queries (${pageTargetInfoId}, ${url})`);
      return ast;
    }).catch(e => {
      cleanupAndExit({
        error: e
      });
    }); // -> [BLOCK FOR] page preparation

    page = await updatedPagePromise;

    if (!page) {
      cleanupAndExit({
        error: 'Could not open page in browser'
      });
      return;
    } // load the page (slow) [NOT BLOCKING]


    const loadPagePromise = loadPage({
      page,
      url,
      pageLoadSkipTimeout,
      allowedResponseCode,
      pageTargetInfoId
    }); // turn css to formatted selectorlist [NOT BLOCKING]

    debuglog(`turn css to formatted selectorlist START (${pageTargetInfoId}, ${url})`);
    const buildSelectorProfilePromise = astPromise.then(astResult => {
      ast = astResult;
      return (0, _selectorsProfile.default)(ast, forceInclude && forceInclude.length ? forceInclude : null, forceExclude && forceExclude.length ? forceExclude : null);
    }).then(res => {
      debuglog(`turn css to formatted selectorlist DONE (${res.selectors.length}) (${pageTargetInfoId}, ${url})`);
      return res;
    }); // -> [BLOCK FOR] page load

    try {
      await loadPagePromise;
    } catch (err) {
      const errorDueToPageUnloaded = PUPPETEER_PAGE_UNLOADED_DURING_EXECUTION_ERROR_REGEX.test(err);
      cleanupAndExit({
        error: errorDueToPageUnloaded ? new Error(PAGE_UNLOADED_DURING_EXECUTION_ERROR_MESSAGE) : err
      });
      return;
    }

    if (!page) {
      // in case we timed out
      cleanupAndExit({
        error: new Error('Page load timed out')
      });
      return;
    }

    if (_hasExited) return;
    let criticalHtmlImagesPromise = Promise.resolve([]);

    if (detectCriticalImages) {
      debuglog(`findCriticalHtmlImages START (${pageTargetInfoId}, ${url})`);
      criticalHtmlImagesPromise = page.evaluate(_criticalHtmlImages.default).then(result => {
        debuglog(`findCriticalHtmlImages DONE (${pageTargetInfoId}, ${url})`);
        return result;
      });
    } // NOTE: cannot safely use page.evaluate before page is available.
    // However it must be ready way before the load event, look into api for this in puppeteer.
    // Also note that this stopAnimationsPromise generally resolves
    // within less than 100ms, the default renderWaitTime with which is runs concurrently.


    let stopAnimationsPromise;

    if (takeScreenshots && screenshots.extraStyles) {
      // so before/after screenshots see the same view:
      stopAnimationsPromise = Promise.all([// run through animations at fast rate
      page._client.send('Animation.setPlaybackRate', {
        playbackRate: 12
      }), // extraStyles can be used for things like:
      // - presenting screenshots on specific background, making custom style changes,
      // disabling animations..
      page.evaluate(_insertStyles.default, screenshots.extraStyles)]).then(() => {
        if (_hasExited) return;
        debuglog(`stopped animations on page, to aid before/after screenshots (${pageTargetInfoId}, ${url})`);
      }).catch(err => {
        debuglog(`stopped animations on page ERROR  ⚠️ (${pageTargetInfoId}, ${url}) - ${err}`);
        throw err;
      });
    } else {
      stopAnimationsPromise = Promise.resolve();
    } // Penthouse waits for the `load` event to fire
    // (before loadPagePromise resolves; except for very slow loading pages)
    // (via default puppeteer page.goto options.waitUntil setting,
    //  https://github.com/GoogleChrome/puppeteer/blob/v1.8.0/docs/api.md#pagegotourl-options)
    // This means "all of the objects in the document are in the DOM, and all the images...
    // have finished loading".
    // This is necessary for Penthouse to know the correct layout of the critical viewport
    // (well really, we would only need to load the critical viewport.. not possible?)
    // renderWaitTime can be used to wait further, but `renderWaitTime` is not a very good name,
    // and just setting a time is also not the most effective solution to f.e. wait for fonts.
    // In future probably deprecate and allow for a custom function instead (returning a promise).
    // (like Puppeteer does)
    // -> [BLOCK FOR] renderWaitTime - needs to be done before we take any screenshots


    try {
      await Promise.all([// also waiting for stop animation promises
      stopAnimationsPromise, new Promise(resolve => {
        setTimeout(() => {
          debuglog(`waited for renderWaitTime:  ${renderWaitTime} (${pageTargetInfoId}, ${url})`);
          resolve();
        }, renderWaitTime);
      })]);
    } catch (err) {
      debuglog(`error thrown, most likely from stopAnimationsPromise (${pageTargetInfoId}, ${url}): ${err}`); // NOTE: if page load was aborted, and no html response was received yet,
      // then this error is _not_ because page was unloaded,
      // but rather because it _had not yet loaded_ - i.e. it's rather a timeout.

      const errorDueToPageUnloaded = PUPPETEER_PAGE_UNLOADED_DURING_EXECUTION_ERROR_REGEX.test(err);
      cleanupAndExit({
        error: errorDueToPageUnloaded ? new Error(PAGE_UNLOADED_DURING_EXECUTION_ERROR_MESSAGE) : err
      });
      return;
    }

    if (_hasExited) return; // take before screenshot (optional) [NOT BLOCKING]

    const beforeScreenshotPromise = takeScreenshots ? grabPageScreenshot({
      type: 'before',
      page,
      screenshots,
      screenshotExtension,
      debuglog,
      pageTargetInfoId,
      url
    }) : Promise.resolve(); // -> [BLOCK FOR] css into formatted selectors list with "sourcemap"
    // latter used to map back to full css rule

    const {
      selectors,
      selectorNodeMap
    } = await buildSelectorProfilePromise;
    if (_hasExited) return;

    if (typeof prepareDomFn === 'function') {
      await page.evaluate(prepareDomFn, {
        width,
        height,
        url
      });
    } // -> [BLOCK FOR] critical css selector pruning (in browser)


    let criticalSelectors;

    try {
      criticalSelectors = await page.evaluate(_pruneNonCriticalSelectors.default, {
        selectors,
        renderWaitTime,
        maxElementsToCheckPerSelector
      }).then(criticalSelectors => {
        debuglog(`pruneNonCriticalSelectors done (${pageTargetInfoId}, ${url})`);
        return criticalSelectors;
      });
    } catch (err) {
      debuglog(`pruneNonCriticalSelector threw an error (${pageTargetInfoId}, ${url}) ${err}`);
      const errorDueToPageUnloaded = PUPPETEER_PAGE_UNLOADED_DURING_EXECUTION_ERROR_REGEX.test(err);
      cleanupAndExit({
        error: errorDueToPageUnloaded ? new Error(PAGE_UNLOADED_DURING_EXECUTION_ERROR_MESSAGE) : err
      });
      return;
    }

    if (_hasExited) return; // -> [BLOCK FOR] clean up final ast for critical css

    debuglog(`AST cleanup START (${pageTargetInfoId}, ${url})`); // NOTE: this function mutates the AST

    (0, _postformatting.default)({
      ast,
      selectorNodeMap,
      criticalSelectors,
      propertiesToRemove,
      maxEmbeddedBase64Length,
      debugContext: `(${pageTargetInfoId}, ${url})`
    });
    debuglog(`AST cleanup DONE (${pageTargetInfoId}, ${url})`); // -> [BLOCK FOR] generate final critical css from critical ast

    let css = _cssTree.default.generate(ast);

    debuglog(`generated CSS from AST (${pageTargetInfoId}, ${url})`);
    let formatCriticalCssPromise = css;

    if (typeof formatCriticalCss === 'function') {
      formatCriticalCssPromise = formatCriticalCss(css);

      if (formatCriticalCssPromise instanceof Promise) {
        formatCriticalCssPromise.then(() => {
          debuglog(`formatCriticalCss DONE (${pageTargetInfoId}, ${url})`);
        });
      } else {
        debuglog(`formatCriticalCss DONE (${pageTargetInfoId}, ${url})`);
      }
    } // take after screenshot (optional) [BLOCKING]


    if (takeScreenshots) {
      try {
        debuglog(`await formatCriticalCss & beforeScreenshot promises (${pageTargetInfoId}, ${url})`) // wait for the before screenshot, before start modifying the page
        ;
        [css] = await Promise.all([formatCriticalCssPromise, beforeScreenshotPromise]);
        if (_hasExited) return;
        debuglog(`inline critical styles for after screenshot (${pageTargetInfoId}, ${url})`);
        await page.evaluate(_replacePageCss.default, {
          css: css + (screenshots.extraStyles || '')
        });
        if (_hasExited) return;
        await grabPageScreenshot({
          type: 'after',
          page,
          screenshots,
          screenshotExtension,
          debuglog,
          pageTargetInfoId,
          url
        });
      } catch (err) {
        debuglog(`error thrown during screenshots (ignoring) (${pageTargetInfoId}, ${url}) ${err}`);
      }
    } else {
      css = await formatCriticalCssPromise;
    }

    const returnValue = {
      css
    };

    if (detectCriticalImages) {
      const criticalHtmlImages = await criticalHtmlImagesPromise;
      detectCriticalImages && debuglog(`findCriticalCssImages START (${pageTargetInfoId}, ${url})`);
      const criticalCssImages = detectCriticalImages ? (0, _criticalCssImages.default)(ast, url) : [];
      detectCriticalImages && debuglog(`findCriticalCssImages DONE (${pageTargetInfoId}, ${url})`);
      returnValue.criticalImages = [].concat(_toConsumableArray(criticalHtmlImages.map(src => ({
        src,
        from: 'html-img-element'
      }))), _toConsumableArray(criticalCssImages.map(src => ({
        src,
        from: 'css-img-url'
      }))));
    }

    debuglog(`generateCriticalCss DONE (${pageTargetInfoId}, ${url})`);
    cleanupAndExit({
      returnValue
    });
  });
}

var _default = pruneNonCriticalCssLauncher;
exports.default = _default;